#!/usr/bin/python3

import socket
import sys
import traceback
from threading import Thread
from time import sleep

class ClunetNotConnectedError(Exception):
    pass

class ClunetAuthError(Exception):
    pass

class ClunetClient():
    """
    Класс для работы с удалённым CLUNET
    """
    PRIO_NOTICE=1
    PRIO_INFO=2
    PRIO_MESSAGE=3
    PRIO_COMMAND=4

    ADDR_SELF=0
    ADDR_LIGHT=1
    ADDR_DOOR_CONTROL=2
    ADDR_RC_RECEIVER=3
    ADDR_RC_TRANSMITTER=4
    ADDR_TEMP_SENSORS=5
    ADDR_INTERCOM=7
    ADDR_MOTION_SENSOR1=8
    ADDR_MOTION_SENSOR2=9
    ADDR_DOOR_LOCK=10
    ADDR_NUMPAD=11
    ADDR_WINDOW=12
    ADDR_DOOR_BELL=13
    ADDR_ELECTRIC_METER=14
    ADDR_BROADCAST=255

    CMD_DISCOVERY=0
    CMD_DISCOVERY_RESP=1
    CMD_BOOT_CONTROL=2
    CMD_REBOOT=3
    CMD_BOOT_COMPLETED=4
    CMD_DOOR_INFO=5
    CMD_DEVICE_POWER_INFO_REQ=6
    CMD_DEVICE_POWER_INFO=7
    CMD_DEVICE_POWER_SET=8
    CMD_SET_TIMER=9
    CMD_RC_BUTTON_PRESSED=10
    CMD_RC_BUTTON_PRESSED_RAW=11
    CMD_RC_BUTTON_SEND=12
    CMD_RC_BUTTON_SEND_RAW=13
    CMD_LIGHT_LEVEL=14
    CMD_ONE_WIRE_START_SEARCH=15
    CMD_ONE_WIRE_DEVICE_FOUND=16
    CMD_TEMPERATURE=17
    CMD_TIME=18
    CMD_WHEEL_INFO=19
    CMD_VOLTAGE=20
    CMD_MOTION=21
    CMD_INTERCOM_RING=22
    CMD_INTERCOM_MESSAGE=23
    CMD_INTERCOM_MODE_REQ=24
    CMD_INTERCOM_MODE_INFO=25
    CMD_INTERCOM_MODE_SET=26
    CMD_INTERCOM_RECORD_REQ=27
    CMD_INTERCOM_RECORD_DATA=28
    CMD_DOOR_LOCK_SET=29
    CMD_DOOR_LOCK_INFO=30
    CMD_DOOR_LOCK_REQ=31
    CMD_BUTTON=32
    CMD_WINDOW_SET=33
    CMD_WINDOW_INFO=34
    CMD_WINDOW_REQ=35
    CMD_DOOR_BELL=36
    CMD_ALARM=37
    CMD_ELECTRO_METER_SET=38
    CMD_ELECTRO_METER_STATS=39
    CMD_ELECTRO_METER_POWER_REQ=40
    CMD_ELECTRO_METER_POWER_INFO=41
    CMD_ELECTRO_METER_STATS_REQ=42
    CMD_PING=254
    CMD_PING_REPLY=255

    def __init__(self, host='localhost', port=10009, password='',
        on_data_received=None, on_connect=None, on_disconnect=None, on_error=None):
        """
        Конструктор. Сразу после создания объекта происходит подключение к серверу.
        Параметры:
            host: адрес сервера
            port: порт сервера
            password: пароль
            on_data_received: функция, которую нужно вызывать при получении данных
            on_connect: функция, которую нужно вызывать при подключении
            on_disconnect: функция, которую нужно вызывать при отключении
        """
        self._host = host
        self._port = port
        self._password = password
        self.on_data_received = on_data_received
        self.on_connect = on_connect
        self.on_disconnect = on_disconnect
        self.on_error = on_error
        self._sock = None        
        self._working = True
        self.connected = False
        self._client_thread = Thread(
            target=self._main_loop,
            name="CLUNET client loop",
            daemon=True)
        self._client_thread.start()
 
    def close(self):
        """
        Разрывает соединение, завершает работу
        """
        self._working = False
        if self._sock:
            self._sock.close()
            self._sock = None

    def _send_line(self, line):        
        self._sock.send((line + "\r\n").encode("ascii"))

    def send(self, prio=PRIO_COMMAND, src=ADDR_SELF, dst=ADDR_BROADCAST, cmd=CMD_DISCOVERY, data=None):
        if not self.connected:
            raise ClunetNotConnectedError("CLUNET is not connected")
        if data == None:
            data = bytes()
        elif type(data) == int:
            data = bytes([data])
        elif type(data) == list:
            data = bytes(data)
        self._send_line("{0:X} {1:02X} {2:02X} {3:02X} {4}"
            .format(prio, src, dst, cmd, data.hex().upper()))

    def _parse_line(self, line):
        line = line.upper()
        if line.startswith("PASSWORD:"):
            self._send_line(self._password)
            return
        if line == "AUTH OK":
            self.connected = True
            if self.on_connect:
                self.on_connect()
            return
        if line == "ACCESS DENIED":
            raise ClunetAuthError("CLUNET - access denied")
        line = "0" + line
        data = bytes.fromhex(line)
        prio = data[0]
        src = data[1]
        dst = data[2]
        cmd = data[3]
        data = data[4:]
        if self.on_data_received:
            self.on_data_received(prio, src, dst, cmd, data)

    def _main_loop(self):
        """
        Внутренний цикл для чтения данных, выполняется фоном в отдельном потоке.
        """
        while self._working:
            buffer = []
            try:
                if not self._sock:
                    self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    self._sock.connect((self._host, self._port))

                while self._sock:
                    data = self._sock.recv(1024)
                    if len(data) > 0:
                        buffer.extend(data)
                    else:
                        break
                    while True:
                        for i in range(len(buffer)):
                            if buffer[i] == 10 or buffer[i] == 13:
                                eol = i
                                break
                        else:
                            break
                        if eol > 0:
                            self._parse_line(bytes(buffer[:eol]).decode('ascii'))
                        buffer = buffer[eol+1:]
            except Exception as e:
                if self.on_error:
                    self.on_error(e)
            finally:
                if self._sock:
                    self._sock.close()
                self._sock = None
                if self.connected and self.on_disconnect:
                    self.on_disconnect()
                self.connected = False
            sleep(5)
