﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Runtime.Serialization.Formatters.Binary;
using System.Text;
using System.Windows.Forms;

namespace Cluster.Clunet.Console
{
    public partial class RcControlForm : Form
    {
        private readonly ClunetConnection clunetConnection;
        private readonly Image rcImage;
        private readonly string configPath;
        private List<RcButton> buttonsMap;
        private RcButton lastButton;
        private DateTime lastSendTime = DateTime.Now;
        double scale = 1;
        public RcControlForm(ClunetConnection clunetConnection, Image rcImage, string configPath, double scale = 1)
        {
            this.clunetConnection = clunetConnection;
            this.rcImage = rcImage;
            this.configPath = configPath;
            InitializeComponent();
            this.Width = rcImage.Width;
            this.Height = rcImage.Height;
            this.scale = scale;
            if (scale != 1)
            {
                var scaled = new Bitmap((int)Math.Round(rcImage.Width * scale), (int)Math.Round(rcImage.Height * scale));
                var gr = Graphics.FromImage(scaled);
                gr.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
                gr.DrawImage(rcImage, new Rectangle(0, 0, scaled.Width, scaled.Height), new Rectangle(0, 0, rcImage.Width, rcImage.Height), GraphicsUnit.Pixel);
                gr.Flush();
                pictureBoxRC.Image = scaled;
            }
            else pictureBoxRC.Image = rcImage;
            LoadConfig();
        }

        public void ScanImage(Image rcMask)
        {
            var buttons = new List<RcButton>();
            for (int y1 = 1; y1 < rcMask.Height; y1++)
                for (int x1 = 1; x1 < rcMask.Width; x1++)
                {
                    if (
                        ((((Bitmap)rcMask).GetPixel(x1, y1).ToArgb() & 0xFFFFFF) == 0x00FF00) &&
                        ((((Bitmap)rcMask).GetPixel(x1, y1 - 1).ToArgb() & 0xFFFFFF) != 0x00FF00) &&
                        ((((Bitmap)rcMask).GetPixel(x1 - 1, y1).ToArgb() & 0xFFFFFF) != 0x00FF00)
                    )
                    {
                        var button = new RcButton();
                        button.top = button.bottom = y1;
                        button.left = button.right = x1;
                        while (
                            ((((Bitmap)rcMask).GetPixel(button.right, button.top).ToArgb() & 0xFFFFFF) == 0x00FF00)
                            && (button.right < rcImage.Width - 1)
                            )
                            button.right++;
                        while (
                            ((((Bitmap)rcMask).GetPixel(button.left, button.bottom).ToArgb() & 0xFFFFFF) == 0x00FF00)
                            && (button.bottom < rcImage.Height - 1)
                            )
                            button.bottom++;
                        button.raw = false;
                        button.code = null;
                        button.key = Keys.None;
                        buttons.Add(button);
                    }
                }
            buttonsMap = buttons;
            SaveConfig();
        }

        private void pictureBoxRC_MouseMove(object sender, MouseEventArgs e)
        {
            if (GetButtonByXY(e.X, e.Y) != null)
                pictureBoxRC.Cursor = Cursors.Hand;
            else pictureBoxRC.Cursor = Cursors.Default;
        }

        private RcButton GetButtonByXY(int x, int y)
        {
            x = (int)Math.Round(x / scale);
            y = (int)Math.Round(y / scale);
            foreach (var button in buttonsMap)
            {
                if (x >= button.left && x <= button.right && y >= button.top && y <= button.bottom)
                    return button;
            }
            return null;
        }

        private RcButton GetButtonByKey(Keys key)
        {
            foreach (var button in buttonsMap)
            {
                if (key == button.key)
                    return button;
            }
            return null;
        }

        private void pictureBoxRC_MouseDown(object sender, MouseEventArgs e)
        {
            var button = GetButtonByXY(e.X, e.Y);
            if (button == null) return;
            lastButton = (RcButton)button;
            if (e.Button == System.Windows.Forms.MouseButtons.Right)
            {
                contextMenuStripButton.Show(System.Windows.Forms.Cursor.Position);
            }
            else if ((e.Button == System.Windows.Forms.MouseButtons.Left)
               && lastButton.code != null)
            {
                try
                {
                    SendButton();
                    timerRepeat.Enabled = true;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(this, ex.Message, "Ошибка", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void pictureBoxRC_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == System.Windows.Forms.MouseButtons.Left)
                timerRepeat.Enabled = false;
        }

        private void SendButton()
        {
            if (lastButton == null) return;
            if (lastButton.code == null) return;
            clunetConnection.Send(new ClunetPacket(ClunetPacket.PacketPriority.Command, ClunetConsoleForm.LocalAddress, 0xFF,
                (!lastButton.raw) ? ClunetPacket.CommandType.RcButtonSend : ClunetPacket.CommandType.RcRawButtonSend,
                lastButton.code));
        }

        private void learnButtonToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lastButton == null) return;
            var form = new RcButtonLearn(clunetConnection, lastButton.code);
            if (form.ShowDialog() == DialogResult.OK)
            {
                lastButton.raw = form.IsRaw;
                lastButton.code = form.ButtonCode;
                SaveConfig();
            }
        }

        private void learnKeyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lastButton == null) return;
            var form = new SelectKeyDialogForm();
            form.ShowDialog();
            lastButton.key = form.Key;
            SaveConfig();
        }

        private void RcControlForm_KeyDown(object sender, KeyEventArgs e)
        {
            if ((DateTime.Now - lastSendTime).TotalMilliseconds < 150) return;
            lastButton = GetButtonByKey(e.KeyCode);
            if (lastButton == null) return;
            timerRepeat.Enabled = false;
            try
            {
                SendButton();
            }
            catch (Exception ex)
            {
                MessageBox.Show(this, ex.Message, "Ошибка", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            lastSendTime = DateTime.Now;
        }

        private void LoadConfig()
        {
            try
            {
                if (File.Exists(configPath))
                {
                    Stream fileStream = File.OpenRead(configPath);
                    BinaryFormatter deserializer = new BinaryFormatter();
                    buttonsMap = (List<RcButton>)deserializer.Deserialize(fileStream);
                    fileStream.Close();
                }
            }
            catch { }
        }

        private void SaveConfig()
        {
            try
            {
                Stream fileStream = File.Create(configPath);
                BinaryFormatter serializer = new BinaryFormatter();
                serializer.Serialize(fileStream, buttonsMap);
                fileStream.Close();
            }
            catch { }
        }

        private void timerRepeat_Tick(object sender, EventArgs e)
        {
            try
            {
                SendButton();
            }
            catch { }
        }

        public void ExportToHtml(string htmlPath, string imageURL, string scriptURL, double scale = 1)
        {
            var file = File.CreateText(htmlPath);
            file.WriteLine(
                "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">");
            file.WriteLine(@"<html>
<head>
<title>Remote Control</title>
<meta name='author' content='Cluster'>
</head>
<body>
<MAP NAME='rcmap'>");
            foreach (var button in buttonsMap)
            {
                if (button.code != null)
                {
                    var code = new StringBuilder();
                    if (!button.raw)
                        code.Append("?code=");
                    else code.Append("?rawcode=");
                    foreach (var b in button.code)
                        code.AppendFormat("{0:X2}", b);
                    file.WriteLine("<AREA SHAPE='RECT' COORDS='" + Math.Round(button.left * scale) + ", " + Math.Round(button.top * scale) + ", " + Math.Round(button.right * scale) + ", " + Math.Round(button.bottom * scale) + "' HREF='" + scriptURL + code.ToString() + "'>");
                }
            }
            file.WriteLine("</map>");
            file.WriteLine("<img ismap usemap='#rcmap' src='" + imageURL + "' border='0'>");
            file.WriteLine("</body></html>");
            file.Close();
        }

        public void ExportToJava(string scriptPath, double scale = 1)
        {
            var file = File.CreateText(scriptPath);
            file.WriteLine("RcButton[] buttonsMap = {");
            bool first = true;
            foreach (var button in buttonsMap)
            {
                if (!first) file.Write(",");
                first = false;
                var code = new StringBuilder();
                if (button.code != null)
                {
                    bool firstByte = true;
                    foreach (var b in button.code)
                    {
                        if (!firstByte) code.Append(",");
                        firstByte = false;
                        code.AppendFormat("(byte)0x{0:X2}", b);
                    }
                }
                file.WriteLine("new RcButton(" + Math.Round(button.left * scale) + ", " + Math.Round(button.top * scale) + ", " + Math.Round(button.right * scale) + ", " + Math.Round(button.bottom * scale) + ", new byte[]{" + code.ToString() + "})");
            }
            file.WriteLine("};");
            file.Close();
        }

        private void RcControlForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.UserClosing)
            {
                e.Cancel = true;
                Hide();
            }
        }
    }


    [Serializable()]
    class RcButton
    {
        public int top, left, bottom, right;
        public bool raw;
        public byte[] code;
        public Keys key;
    }
}
