﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace Cluster.Clunet
{
    public class ClunetConnection : IDisposable
    {
        public delegate void PacketReceivedDelegate(ClunetPacket packet);
        public delegate void CrcErrorDelegate();
        public delegate void ConnectedDelegate();
        public delegate void DisconnectedDelegate();
        public event PacketReceivedDelegate OnPacketReceived = delegate { };
        public event ConnectedDelegate OnConnected = delegate { };
        public event DisconnectedDelegate OnDisconnected = delegate { };
        private Socket socket = null;
        private Thread mainThread = null;
        private Thread connectThread = null;
        private readonly string host;
        private readonly int port;
        private readonly string password;
        private bool autoconnect = false;
        private bool authOk = false;

        public bool Autoconnect
        {
            get { return autoconnect; }
            set
            {
                if (autoconnect == value) return;
                autoconnect = value;
                if (autoconnect)
                {
                    connectThread = new Thread(ConnectThread);
                    connectThread.Start();
                }
                else
                {
                    connectThread.Abort();
                    connectThread = null;
                }
            }
        }

        public ClunetConnection(string host, int port, string password)
        {
            this.host = host;
            this.port = port;
            this.password = password;
        }

        public void Connect()
        {
            if (mainThread != null)
            {
                try
                {
                    mainThread.Abort();
                }
                catch { }
            }
            if (socket != null)
            {
                try
                {
                    socket.Close();
                }
                catch { }
            }
            socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            authOk = false;
            socket.Connect(host, port);
            mainThread = new Thread(MainThread);
            mainThread.Start();
            OnConnected();
        }

        public void Disconnect()
        {
            if (mainThread != null)
            {
                try
                {
                    mainThread.Abort();
                }
                catch { }
            }
            if (socket != null)
            {
                try
                {
                    socket.Close();
                    socket = null;
                }
                catch { }
            }
        }

        private void NewLine(string line)
        {
            if (line.ToLower().StartsWith("password:"))
            {
                try
                {
                    socket.Send(Encoding.ASCII.GetBytes(password + "\n"));
                }
                catch { }
            }
            else if (line.ToLower().StartsWith("auth ok"))
            {
                if (!authOk)
                {
                    authOk = true;
                    OnConnected();
                }
            }
            else if (line.ToLower().StartsWith("access denied"))
            {
                authOk = false;
                Autoconnect = false;
            }
            else
            {
                var packet = new ClunetPacket(line);
                OnPacketReceived(packet);
            }
        }

        private void MainThread()
        {
            try
            {
                var buf = new byte[1024];
                int pos = 0;
                //socket.ReceiveTimeout = 60000;
                while (true)
                {
                    if (socket.Poll(60000000, SelectMode.SelectRead))
                    {
                        var res = socket.Receive(buf, pos, 1, SocketFlags.None);
                        if (res > 0)
                        {
                            if ((buf[pos] == '\r') || (buf[pos] == '\n'))
                            {
                                if (pos > 0)
                                    NewLine(Encoding.ASCII.GetString(buf, 0, pos));
                                pos = 0;
                            }
                            else pos += res;
                        }
                        if (res == 0)
                        {
                            socket.Close();
                            socket = null;
                            OnDisconnected();
                            return;
                        }
                        if (pos >= buf.Length) pos = 0;
                    }
                    else throw new SocketException();
                }
            }
            catch (SocketException)
            {
                try
                {
                    socket.Close();
                    OnDisconnected();
                }
                catch { }
            }
            catch (ThreadAbortException) { }
        }

        public void Send(ClunetPacket packet)
        {
            if (!Connected) throw new NotConnectedException();
            string line = packet.CommandToSend + "\r\n";
            socket.Send(Encoding.ASCII.GetBytes(line));
        }

        public bool Connected
        {
            get
            {
                return (socket != null) && (socket.Connected) && authOk;
            }
        }

        private void ConnectThread()
        {
            try
            {
                while (autoconnect)
                {
                    if (!Connected)
                    {
                        try
                        {
                            Connect();
                        }
                        catch (ThreadAbortException)
                        {
                            return;
                        }
                        catch (Exception) { }
                    }
                    Thread.Sleep(5000);
                }
            }
            catch (ThreadAbortException) { }
        }

        public void Dispose()
        {
            Autoconnect = false;
            Disconnect();
        }
    }

    public class NotConnectedException : Exception
    {
    }
}
