﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;

namespace Cluster.Clunet
{
    public class ClunetPacket
    {
        public enum PacketPriority { Notice = 1, Info = 2, Message = 3, Command = 4 };
        public enum CommandType
        {
            Discovery = 0x00,
            DiscoveryResponse = 0x01,
            BootControl = 0x02,
            Reboot = 0x03,
            BootCompleted = 0x04,
            DoorInfo = 0x05,
            DevicePowerInfoRequest = 0x06,
            DevicePowerInfo = 0x07,
            DevicePowerCommand = 0x08,
            SetTimer = 0x09,
            RcButtonPressed = 0x0A,
            RcRawButtonPressed = 0x0B,
            RcButtonSend = 0x0C,
            RcRawButtonSend = 0x0D,
            LightLevel = 0x0E,
            OneWireStartSearch = 0x0F,
            OneWireDeviceFound = 0x10,
            Temperature = 0x11,
            Time = 0x12,
            WheelInfo = 0x13,
            Voltage = 0x14,
            Motion = 0x15,
            IntercomRing = 0x16,
            IntercomMessage = 0x17,
            IntercomModeRequest = 0x18,
            IntercomModeInfo = 0x19,
            IntercomModeSet = 0x1A,
            IntercomRecordRequest = 0x1B,
            IntercomRecordData = 0x1C,
            DoorLockCommand = 0x1D,
            DoorLockInfo = 0x1E,
            DoorLockInfoRequest = 0x1F,
            Button = 0x20,
            WindowCommand = 0x21,
            WindowInfo = 0x22,
            WindowInfoRequest = 0x23,
            DoorBell = 0x24,
            Alarm = 0x25,
            ElectricMeterSet = 0x26,
            ElectricMeterStats = 0x27,
            ElectcicMeterPowerRequest = 0x28,
            ElectricMeterPowerInfo = 0x29,
            ElectricMeterStatsRequest = 0x30,
            //HammerConfigSet = 0x80,
            //HammerConfigInfo = 0x81,
            //HammerStateSet = 0x82,
            //HammerStateInfo = 0x83,
            Ping = 0xFE,
            PingReply = 0xFF
        };

        public ClunetPacket(string hexData)
        {
            hexData = hexData.Replace(" ", "");
            if ((hexData.Length - 1) / 2 < 3) throw new InvalidPacketDataException();
            Priority = (PacketPriority)(hexData[0] - '0');
            var buffer = new byte[(hexData.Length - 1) / 2];
            try
            {
                for (int i = 0; i < hexData.Length - 1; i += 2)
                {
                    buffer[i / 2] = byte.Parse(hexData.Substring(1 + i, 2), NumberStyles.AllowHexSpecifier);
                }
            }
            catch (Exception)
            {
                throw new InvalidPacketDataException();
            }

            SrcAddress = buffer[0];
            DstAddress = buffer[1];
            Command = buffer[2];
            Data = new byte[buffer.Length - 3];
            Array.Copy(buffer, 3, Data, 0, buffer.Length - 3);
        }

        public ClunetPacket(PacketPriority priority, byte srcAddress, byte dstAddress, byte command, byte[] data)
        {
            this.Priority = priority;
            this.SrcAddress = srcAddress;
            this.DstAddress = dstAddress;
            this.Command = command;
            if (data != null)
            {
                this.Data = new byte[data.Length];
                Array.Copy(data, this.Data, data.Length);
            }
            else
            {
                this.Data = new byte[0];
            }
        }
        public ClunetPacket(PacketPriority priority, byte srcAddress, byte dstAddress, byte command)
            : this(priority, srcAddress, dstAddress, command, null) { }
        public ClunetPacket(PacketPriority priority, byte srcAddress, byte dstAddress, CommandType command, byte[] data)
            : this(priority, srcAddress, dstAddress, (byte)command, data) { }
        public ClunetPacket(PacketPriority priority, byte srcAddress, byte dstAddress, CommandType command)
            : this(priority, srcAddress, dstAddress, (byte)command) { }

        public string CommandToSend
        {
            get
            {
                var result = new StringBuilder();
                result.AppendFormat("{0:X1} {1:X2} {2:X2} {3:X2}", (int)Priority, SrcAddress, DstAddress, Command);
                var data = Data;
                foreach (var b in data) result.AppendFormat("{0:X2}", b);
                return result.ToString();
            }
        }

        public byte SrcAddress { get; set; }
        public PacketPriority Priority { get; set; }
        public byte DstAddress { get; set; }
        public byte Command { get; set; }
        public byte Length
        {
            get
            {
                return (byte)Data.Length;
            }
        }
        public byte[] Data { get; set; }
        public string DataAsString
        {
            get
            {
                return Encoding.UTF8.GetString(Data);
            }
        }

        public override string ToString()
        {
            return string.Format("SRC:{0:X2} DST:{1:X2} CMD:{2:X2}({3}) LEN:{4:X2} DATA:({5})", SrcAddress, DstAddress, Command, (CommandType)Command, Length, DataAsHex);
        }

        public string DataAsHex
        {
            get
            {
                var result = new StringBuilder();
                foreach (var b in Data) result.AppendFormat("{0:X2}", b);
                return result.ToString();
            }
        }

        public string HumanReadableInfo()
        {
            try
            {
                string str1 = "???";
                switch ((CommandType)Command)
                {
                    case CommandType.Discovery:
                        return "Поиск устройств в сети";
                    case CommandType.DiscoveryResponse:
                        return "Ответ на поиск, имя устройства: " + DataAsString;
                    case CommandType.BootControl:
                        switch (Data[0])
                        {
                            case 0: str1 = "загрузчик запущен"; break;
                            case 1: str1 = "перейти в режим обновления прошивки"; break;
                            case 2: str1 = "перешёл в режим прошивки, размер страницы " + (Data[1] + Data[2] * 0x100) + " байт"; break;
                            case 3: str1 = "запись прошивки по адресу " + string.Format("{1:X2}{0:X2}", Data[1], Data[2]) + " байт"; break;
                            case 4: str1 = "блок записан"; break;
                            case 5: str1 = "выйти из режима прошивки"; break;
                        }
                        return "Работа с загрузчиком: " + str1;
                    case CommandType.Reboot:
                        return "Перезагрузись!";
                    case CommandType.BootCompleted:
                        switch (Data[0])
                        {
                            case 1: str1 = "холодный старт"; break;
                            case 2: str1 = "внешний ресет"; break;
                            case 4: str1 = "brown-out ресет"; break;
                            case 8: str1 = "перезагружено по команде или WDT"; break;
                            default: str1 = string.Format("??? ({0:X2})", Data[0]); break;
                        }
                        return "Устройство запущено: " + str1;
                    case CommandType.DoorInfo:
                        switch (Data[0])
                        {
                            case 0: str1 = "закрыта"; break;
                            case 1: str1 = "открыта"; break;
                            case 2: str1 = "тревога"; break;
                        }
                        return "Состояние двери: " + str1;
                    case CommandType.DevicePowerInfoRequest:
                        return "Запрос состояния реле";
                    case CommandType.DevicePowerInfo:
                        return string.Format("Состояние реле: {0:X2}", Data[0]);
                    case CommandType.DevicePowerCommand:
                        switch (Data[0])
                        {
                            case 0xFE: str1 = "убавить"; break;
                            case 0xFF: str1 = "прибавить"; break;
                            default: str1 = string.Format("{0:X2}", Data[0]); break;
                        }
                        return "Изменить состояние реле: " + str1;
                    case CommandType.SetTimer:
                        return "Установить таймер на " + (Data[0] + Data[1] * 0x100) + " секунд";
                    case CommandType.RcButtonPressed:
                        if (Data[0] % 2 == 1)
                            return string.Format("Нажата кнопка на пульте ДУ, код: {0:X2}{1:X2}{2:X2}{3:X2}", Data[4], Data[3], Data[2], Data[1]);
                        else
                            return string.Format("Удерживается кнопка на пульте ДУ, код: {0:X2}{1:X2}{2:X2}{3:X2}", Data[4], Data[3], Data[2], Data[1]);
                    case CommandType.RcRawButtonPressed:
                        return "Нажата кнопка на пульте ДУ, декодировать не удалось, в пакете сырые данные";
                    case CommandType.RcButtonSend:
                        //if (data[0] == 1)
                        return string.Format("Заэмулировать кнопку на пульте ДУ, код: {0:X2}{1:X2}{2:X2}{3:X2}, удерживать {4} п.", Data[4], Data[3], Data[2], Data[1], ((Length < 6) ? 0 : Data[5]));
                    //if (data[0] == 2)
                    //  return "Заэмулировать удержание кнопки на пульте ДУ";
                    case CommandType.RcRawButtonSend:
                        return "Заэмулировать кнопку на пульте ДУ на основе сырых данных";
                    case CommandType.LightLevel:
                        return "Уровень освещения: " + (Data[0] + Data[1] * 0x100);
                    case CommandType.OneWireStartSearch:
                        return "Начать поиск 1-wire устройств";
                    case CommandType.OneWireDeviceFound:
                        return string.Format("Найдено 1-wire устройство: тип - {0:X2}, серийный номер - {6:X2}{5:X2}{4:X2}{3:X2}{2:X2}{1:X2}, CRC - {7:X2}", Data[0], Data[1], Data[2], Data[3], Data[4], Data[5], Data[6], Data[7]);
                    case CommandType.Temperature:
                        return string.Format("Температура на датчике {0:X2}-{6:X2}{5:X2}{4:X2}{3:X2}{2:X2}{1:X2}: {7} °C", Data[0], Data[1], Data[2], Data[3], Data[4], Data[5], Data[6], GetTemperature());
                    case CommandType.Time:
                        return string.Format("Время: {5:D2}.{4:D2}.20{3:D2} {0:D2}:{1:D2}:{2:D2} ", Data[0], Data[1], Data[2], Data[3] % 100, Data[4], Data[5]);
                    case CommandType.WheelInfo:
                        return string.Format("Обороты колеса #{0}: {1}", Data[0], Data[1] + Data[2] * 256);
                    case CommandType.Voltage:
                        return string.Format("Устройство #{0}, напряжение на батарейке: {1} вольт", Data[0], 2.56 * 2 * (Data[1] + Data[2] * 256) / 1024.0);
                    case CommandType.Motion:
                        return string.Format("Обнаружено движение на датчике #{0}", Data[0]);
                    case CommandType.IntercomRing:
                        return "Звонок в домофон, используется режим номер " + Data[0];
                    case CommandType.IntercomMessage:
                        return string.Format("Через домофон оставлено сообщение номер {0}", Data[0] + Data[1] * 0x100 + Data[2] * 0x10000 + Data[3] * 0x1000000);
                    case CommandType.IntercomModeRequest:
                        return "Запрос текущего режима домофона";
                    case CommandType.IntercomModeInfo:
                        return string.Format("Режим домофона. Постоянный: {0}. Разовый: {1}.", ((Data[0] == 0) ? "выключен" : Data[0].ToString()), ((Data[1] == 0) ? "отсутствует" : Data[1].ToString()));
                    case CommandType.IntercomModeSet:
                        return string.Format("Установка режима домофона. Постоянный: {0}. Разовый: {1}.", ((Data[0] == 0xFF) ? "не трогать" : Data[0].ToString()), ((Data.Length < 2 || Data[1] == 0xFF) ? "не трогать" : Data[1].ToString()));
                    case CommandType.IntercomRecordRequest:
                        if (Data.Length == 4)
                            return string.Format("Запрошена передача сообщения номер {0}", Data[0] + Data[1] * 0x100 + Data[2] * 0x10000 + Data[3] * 0x1000000);
                        else if (Data.Length == 1)
                        {
                            if (Data[0] == 1) return "Подтверждён приём записи с домофона";
                            if (Data[0] == 0) return "Завершение передачи записи с домофона";
                        }
                        break;
                    case CommandType.IntercomRecordData:
                        return string.Format("Передача сообщения номер {0}, смещение {1:X8}", Data[0] + Data[1] * 0x100 + Data[2] * 0x10000 + Data[3] * 0x1000000,
                            Data[4] + Data[5] * 0x100 + Data[6] * 0x10000 + Data[7] * 0x1000000);
                    case CommandType.DoorLockCommand:
                        if (Data[0] == 1) return "Открыть дверной замок";
                        if (Data[0] == 2) return "Закрыть дверной замок";
                        break;
                    case CommandType.DoorLockInfo:
                        if (Data[0] == 1) return "Открыт дверной замок";
                        if (Data[0] == 2) return "Закрыт дверной замок";
                        break;
                    case CommandType.DoorLockInfoRequest:
                        return "Запрос состояния дверного замка";
                    case CommandType.Button:
                        return "Нажата кнопка: " + DataAsHex;
                    case CommandType.WindowCommand:
                        if (Data[0] == 0) return "Открыть окно полностью";
                        if (Data[0] == 1) return "Открыть окно наполовину ";
                        if (Data[0] == 2) return "Закрыть окно";
                        if (Data[0] == 3) return "Открыть окно чуть-чуть";
                        if (Data[0] == 4) return "Закрыть окно чуть-чуть";
                        break;
                    case CommandType.WindowInfo:
                        if (Data[0] == 0) return "Окно открыто полностью";
                        if (Data[0] == 1) return "Окно открыто наполовину";
                        if (Data[0] == 2) return "Окно закрыто";
                        if (Data[0] == 0x7F) return "Окно в кастомном состоянии";
                        break;
                    case CommandType.WindowInfoRequest:
                        return "Запрос состояния окна";
                    case CommandType.Ping:
                        return string.Format("Пинг размером в {0} байт", Data.Length);
                    case CommandType.PingReply:
                        return "Ответ на пинг";
                    case CommandType.DoorBell:
                        return "Звонок в дверь";
                    case CommandType.Alarm:
                        return string.Format("Тревога, звук номер {0}", Data[0]);
                    case CommandType.ElectricMeterSet:
                        if (Data.Length >= 6)
                            return string.Format("Установить показания электросчётчика: T1={0}, T2={1}",
                                Data[0] + Data[1] * 0x100 + Data[2] * 0x10000 + Data[3] * 0x1000000 +
                                (Data[4] + Data[5] * 0x100) / 10000.0,
                                Data.Length > 6 ?
                                (Data[6] + Data[7] * 0x100 + Data[8] * 0x10000 + Data[9] * 0x1000000 +
                                (Data[10] + Data[11] * 0x100) / 10000.0) : 0);
                        else return "Сохранить показания электросчётчика";
                    case CommandType.ElectricMeterStats:
                        return string.Format("Текущие показания электросчётчика: T1={0}, T2={1}",
                            Data[0] + Data[1] * 0x100 + Data[2] * 0x10000 + Data[3] * 0x1000000 +
                            (Data[4] + Data[5] * 0x100) / 10000.0,
                            Data[6] + Data[7] * 0x100 + Data[8] * 0x10000 + Data[9] * 0x1000000 +
                            (Data[10] + Data[11] * 0x100) / 10000.0);
                    case CommandType.ElectcicMeterPowerRequest:
                        return "Запрос нагрузки электросчётчика";
                    case CommandType.ElectricMeterPowerInfo:
                        return string.Format("Текущая электронрагрузка: {0} Вт",
                            720000 / (Data[0] + Data[1] * 0x100));
                    case CommandType.ElectricMeterStatsRequest:
                        return "Запрос показаний электросчётчика";
                }
                return "???";
            }

            catch (Exception)
            {
                return "Ошибка декодирования";
            }
        }

        public double GetTemperature()
        {
            if ((CommandType)Command != CommandType.Temperature) throw new Exception("Это не пакет с температурой");
            double temperature = 0;
            if (Data[0] == 0x10)
                temperature = Data[7] / 2.0;
            else if (Data[0] == 0x28)
                temperature = (Data[7] + ((Data[8] & 0x07) << 8)) / 16.0;
            else throw new Exception("Это не термодатчик или неизвестный тип датчика");
            if (Data[8] >> 7 == 1) temperature -= 128;
            return temperature;
        }
    }

    public class InvalidPacketDataException : Exception
    {
    }
}
